"""Stormgate replay parsing tools"""
import gzip
import sys
from contextlib import contextmanager
from pathlib import Path
from pydantic import BaseModel
from typing import BinaryIO, Iterable, List, Optional, Union

from .stormgate_pb2 import ReplayChunk


@contextmanager
def decompress(replay: Union[Path, BinaryIO]):
    """Open a gzipped stormgate replay, skipping the 16-byte header."""
    if isinstance(replay, Path):
        replay = replay.open("rb")
    with replay:
        replay.seek(16)
        with gzip.GzipFile(fileobj=replay) as f2:
            yield f2


def read_varint(f) -> Optional[int]:
    """Read a base-7 varint from a binary stream"""
    bs = f.read(1)
    if len(bs) == 0:
        return None
    byte = ord(bs)
    digits = byte & 0b01111111
    if byte & 0b10000000:
        nxt = read_varint(f)
        if nxt is None:
            raise ValueError("EOF encountered while parsing varint")
        digits += nxt << 7
    return digits


def split_replay(replay: Union[Path, BinaryIO]) -> Iterable[bytes]:
    """Split a replay into a sequence of chunks, each of which is a raw
    bytestring containing a wire-format encoding of a protobuf message."""
    with decompress(replay) as f:
        while True:
            length = read_varint(f)
            if length is None:
                break
            yield f.read(length)


class MatchInfo(BaseModel):
    player_nicknames: List[str] = []
    map_name: Optional[str] = None


def get_match_info(replay: Union[Path, BinaryIO]) -> MatchInfo:
    """Parse what we can from a stormgate replay."""
    info = MatchInfo()
    for bytestring in split_replay(replay):
        chunk = ReplayChunk.FromString(bytestring)
        content = chunk.inner.content
        content_type = content.WhichOneof("contenttype")
        if content_type == "map":
            info.map_name = content.map.name
        if content_type == "player":
            info.player_nicknames.append(content.player.name.nickname)
        if len(info.player_nicknames) >= 2 and info.map_name is not None:
            break
    return info



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Map(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class Player(_message.Message):
    __slots__ = ("name",)
    class PlayerName(_message.Message):
        __slots__ = ("nickname", "numbers")
        NICKNAME_FIELD_NUMBER: _ClassVar[int]
        NUMBERS_FIELD_NUMBER: _ClassVar[int]
        nickname: str
        numbers: str
        def __init__(self, nickname: _Optional[str] = ..., numbers: _Optional[str] = ...) -> None: ...
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: Player.PlayerName
    def __init__(self, name: _Optional[_Union[Player.PlayerName, _Mapping]] = ...) -> None: ...

class ReplayChunk(_message.Message):
    __slots__ = ("inner",)
    class Wrapper(_message.Message):
        __slots__ = ("content",)
        class ReplayContent(_message.Message):
            __slots__ = ("map", "player")
            MAP_FIELD_NUMBER: _ClassVar[int]
            PLAYER_FIELD_NUMBER: _ClassVar[int]
            map: Map
            player: Player
            def __init__(self, map: _Optional[_Union[Map, _Mapping]] = ..., player: _Optional[_Union[Player, _Mapping]] = ...) -> None: ...
        CONTENT_FIELD_NUMBER: _ClassVar[int]
        content: ReplayChunk.Wrapper.ReplayContent
        def __init__(self, content: _Optional[_Union[ReplayChunk.Wrapper.ReplayContent, _Mapping]] = ...) -> None: ...
    INNER_FIELD_NUMBER: _ClassVar[int]
    inner: ReplayChunk.Wrapper
    def __init__(self, inner: _Optional[_Union[ReplayChunk.Wrapper, _Mapping]] = ...) -> None: ...



"""Rename stormgate replays to include useful info in filename"""
from datetime import datetime, timedelta, timezone
import json
from pathlib import Path
import re
import logging
from shutil import copytree
from time import sleep
from typing import NamedTuple
from urllib.request import urlopen, Request

import pandas as pd

from shroudstone.replay import get_match_info
from shroudstone import config, __version__

logger = logging.getLogger(__name__)

STORMGATEWORLD = "https://api.stormgateworld.com"
TOLERANCE = timedelta(seconds=90)
"""Maximum time difference to consider games a match"""
BAD_CHARS = re.compile(r'[<>:"/\\|?*\0]')
"""Characters forbidden in filenames on Linux or Windows"""
USER_AGENT = f"shroudstone v{__version__}"
"""User-Agent to use in requests to the API"""


def rename_replays(
    replay_dir: Path,
    my_player_id: str,
    dry_run: bool = False,
    backup: bool = True,
    reprocess: bool = False,
    check_nicknames: bool = True,
):
    if dry_run:
        # Don't bother
        bu_dir = None
        logger.warning(
            "Performing a dry run - will show what would happen but not actually touch anything."
        )
    elif backup:
        bu_dir = replay_dir.parent / f"{replay_dir.name}.backup"
        backup_dir(replay_dir, bu_dir)
    else:
        bu_dir = None
        logger.warning("No backup being made! You asked for it!")

    if reprocess:
        # Reprocess all replays
        pattern = "**/*.SGReplay"
        logger.info(f"Searching for all replays in {replay_dir}.")
    else:
        # Only look for replays we haven't already renamed
        pattern = "**/CL*.SGReplay"
        logger.info(f"Searching for unrenamed replays in {replay_dir}.")

    replays = [
        x for x in map(ReplayFile.from_path, replay_dir.glob(pattern)) if x is not None
    ]
    if not replays:
        logger.warning(
            "No replays found to rename! "
            f"If you weren't expecting this, check your replay_dir '{replay_dir}' is correct."
        )
        return

    n = len(replays)
    earliest_time = min(x.time for x in replays)
    logger.info(
        f"Found {n} unrenamed replays going back to {earliest_time}. Fetching matches from stormgateworld now."
    )
    matches = player_matches_since(my_player_id, earliest_time)

    for r in replays:
        match = find_match(matches, r, check_nicknames=check_nicknames)
        if match is not None:
            try:
                rename_replay(r, match, dry_run=dry_run)
            except Exception as e:
                logger.error(f"Unexpected error handling {r.path}: {e}")
        else:
            logger.error(
                f"No match found for {r.path.name} in stormgateworld match history."
            )

    if bu_dir:
        logger.warning(
            f"Renaming completed. Your replays were backed up to {bu_dir}. "
            "Please check that nothing went wrong now - subsequent runs will overwrite your backup!"
        )


def backup_dir(replay_dir: Path, bu_dir: Path):
    logger.warning(
        f"Backing up your replays to {bu_dir}. "
        "Please check the results after this run - subsequent runs will overwrite your backup! "
    )
    copytree(replay_dir, bu_dir, dirs_exist_ok=True)


def naive_localtime_to_utc(dt: datetime) -> datetime:
    assert dt.tzinfo is None
    return dt.astimezone(timezone.utc).replace(tzinfo=None)


class ReplayFile(NamedTuple):
    path: Path
    time: datetime

    @staticmethod
    def from_path(path: Path):
        # Original names use local times:
        if m := re.search(r"(\d\d\d\d)\.(\d\d)\.(\d\d)-(\d\d).(\d\d)", path.name):
            time = naive_localtime_to_utc(
                datetime(*(int(x) for x in m.groups()))  # type: ignore
            )
        # Our renamed versions use UTC:
        elif m := re.search(r"(\d\d\d\d)-(\d\d)-(\d\d) (\d\d).(\d\d)", path.name):
            time = datetime(*(int(x) for x in m.groups()))  # type: ignore
        else:
            return None
        return ReplayFile(path, time)


def player_matches_since(player_id: str, time: datetime):
    page = 1
    matches = player_matches(player_id, page=page)
    if matches is None:
        raise Exception(
            "Could not find any matches for {player_id=} - check this is correct!"
        )
    while matches.created_at.min() > time:
        sleep(1)  # Be nice to the API servers! You can wait a few seconds!
        logger.info(f"Got matches back to {matches.created_at.min()}, going further.")
        page += 1
        next_page = player_matches(player_id, page=page)
        if next_page is None:
            logger.warning(
                f"Could only fetch back to {matches.created_at.min()} "
                "from stormgateworld but you have unrenamed replays from before this.",
            )
            return matches.reset_index(drop=True)
        matches = pd.concat([matches, next_page])
    return matches.reset_index(drop=True)


def player_matches(player_id: str, page: int):
    logger.info(f"Fetching match history page {page}...")
    request = Request(
        f"{STORMGATEWORLD}/v0/players/{player_id}/matches?page={page}",
        headers={"User-Agent": USER_AGENT},
    )
    with urlopen(request) as f:
        data = json.load(f)["matches"]
    if not data:
        return None
    data = [flatten_match(player_id, x) for x in data]
    matches = pd.json_normalize([x for x in data if x is not None])
    matches["ended_at"] = pd.to_datetime(matches["ended_at"])
    matches["created_at"] = pd.to_datetime(matches["created_at"])
    return matches


def flatten_match(player_id: str, match: dict):
    """Rewrite the players array as two keys `us` and `them`."""
    try:
        match["us"] = next(
            p for p in match["players"] if p["player"]["player_id"] == player_id
        )
        match["them"] = next(
            p for p in match["players"] if p["player"]["player_id"] != player_id
        )
    except Exception:
        return None
    del match["players"]
    return match


def find_match(matches: pd.DataFrame, replay: ReplayFile, check_nicknames: bool):
    """Given a replay file and a list of matches from stormgateworld, find the
    closest match in time with the correct nicknames."""
    delta = (matches.created_at - replay.time).abs()
    df = matches.assign(delta=delta)
    candidates = df[delta < TOLERANCE].sort_values("delta")
    for _, match in candidates.iterrows():
        info = get_match_info(replay.path)
        try:
            match["map_name"] = info.map_name
        except Exception:
            pass

        try:
            n1 = match["us.player.nickname"]
            n2 = match["them.player.nickname"]
        except Exception:
            n1 = "?"
            n2 = "?"

        if not check_nicknames:
            return match

        nickname_diff = {n1, n2}.symmetric_difference(info.player_nicknames)

        if not nickname_diff:
            return match
        else:
            ns1 = sorted([n1, n2])
            ns2 = sorted(info.player_nicknames)
            logger.error(
                f"Found time-based match for {replay.path.name} "
                f"but nicknames didn't match: {ns1} != {ns2}. "
                "Provide --no-check-nicknames to turn off this check and accept this as a match anyway."
            )


def rename_replay(replay: ReplayFile, match: pd.Series, dry_run: bool):
    us = match["us.player.nickname"]
    them = match["them.player.nickname"]

    try:
        r1 = match["us.race"][0].upper()
    except Exception:
        r1 = "?"

    try:
        r2 = match["them.race"][0].upper()
    except Exception:
        r2 = "?"

    try:
        result = match["us.result"][0].upper()
    except Exception:
        result = "?"

    try:
        map_name = match["map_name"]
    except Exception:
        map_name = "Unknown Map"

    try:
        minutes, seconds = divmod(int(match["duration"]), 60)
        duration = f"{minutes:02d}m{seconds:02d}s"
        time = match["created_at"].strftime("%Y-%m-%d %H.%M")
    except Exception:
        duration = "??m??s"
        time = replay.time.strftime("%Y-%m-%d %H.%M")

    try:
        newname = (
            f"{time} {result} {duration} {us} {r1}v{r2} {them} - {map_name}.SGReplay"
        )
        target = replay.path.parent / newname
        do_rename(replay.path, target, dry_run=dry_run)
    except Exception as e:
        logger.error(f"Unexpected error renaming {replay.path}: {e}")


def do_rename(source: Path, target: Path, dry_run: bool):
    if target.exists():
        logger.error(f"Not renaming {source}! {target} already exists!")
        return

    if dry_run:
        logger.info(f"DRY RUN: Would have renamed {source.name} => {target.name}.")
        return

    logger.info(f"Renaming {source.name} => {target.name}.")
    try:
        source.rename(target)
    except Exception as e:
        # In case the error was due to weird characters in a player name:
        new_name = sanitize_filename(target.name)
        if new_name != target.name:
            logger.warning(
                f"Error renaming {source} => {target.name}, retrying with sanitized filename."
            )
            do_rename(source, target.parent / new_name, dry_run=dry_run)
        else:
            logger.error(f"Error renaming {source} => {target.name}: {e}")


def sanitize_filename(filename: str) -> str:
    """Remove bad characters from a filename"""
    return BAD_CHARS.sub("", filename)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: stormgate.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fstormgate.proto\x12\tstormgate\"\x13\n\x03Map\x12\x0c\n\x04name\x18\x02 \x01(\t\"e\n\x06Player\x12*\n\x04name\x18\x03 \x01(\x0b\x32\x1c.stormgate.Player.PlayerName\x1a/\n\nPlayerName\x12\x10\n\x08nickname\x18\x01 \x01(\t\x12\x0f\n\x07numbers\x18\x02 \x01(\t\"\xeb\x01\n\x0bReplayChunk\x12-\n\x05inner\x18\x03 \x01(\x0b\x32\x1e.stormgate.ReplayChunk.Wrapper\x1a\xac\x01\n\x07Wrapper\x12=\n\x07\x63ontent\x18\x01 \x01(\x0b\x32,.stormgate.ReplayChunk.Wrapper.ReplayContent\x1a\x62\n\rReplayContent\x12\x1d\n\x03map\x18\x03 \x01(\x0b\x32\x0e.stormgate.MapH\x00\x12#\n\x06player\x18\x0c \x01(\x0b\x32\x11.stormgate.PlayerH\x00\x42\r\n\x0b\x63ontenttypeb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'stormgate_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_MAP']._serialized_start=30
  _globals['_MAP']._serialized_end=49
  _globals['_PLAYER']._serialized_start=51
  _globals['_PLAYER']._serialized_end=152
  _globals['_PLAYER_PLAYERNAME']._serialized_start=105
  _globals['_PLAYER_PLAYERNAME']._serialized_end=152
  _globals['_REPLAYCHUNK']._serialized_start=155
  _globals['_REPLAYCHUNK']._serialized_end=390
  _globals['_REPLAYCHUNK_WRAPPER']._serialized_start=218
  _globals['_REPLAYCHUNK_WRAPPER']._serialized_end=390
  _globals['_REPLAYCHUNK_WRAPPER_REPLAYCONTENT']._serialized_start=292
  _globals['_REPLAYCHUNK_WRAPPER_REPLAYCONTENT']._serialized_end=390
# @@protoc_insertion_point(module_scope)



"""The shroudstone CLI"""

import logging
import os
from pathlib import Path
import platform
from typing import Annotated, Optional

from rich.console import Console
from rich.logging import RichHandler
import typer

from shroudstone import renamer, replay
from shroudstone.config import Config

app = typer.Typer()

console = Console(stderr=True)
logging.captureWarnings(True)
logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(console=console)],
)
logger = logging.getLogger(__name__)


@app.command()
def get_replay_info(replay_file: typer.FileBinaryRead):
    """Extract information from a replay, outputting it in JSON format."""
    typer.echo(replay.get_match_info(replay_file).model_dump_json(indent=2))


@app.command()
def split_replay(replay_file: typer.FileBinaryRead, output_directory: Path):
    """Extract a stormgate replay into a directory containing individual protoscope messages."""
    output_directory.mkdir(exist_ok=True, parents=True)
    i = 0
    for i, chunk in enumerate(replay.split_replay(replay_file)):
        (output_directory / f"{i:07d}.binpb").write_bytes(chunk)
    typer.echo(
        f"Wrote {i+1} replay messages in protoscope wire format to {output_directory}/."
    )


@app.command()
def create_rename_replays_shortcut():
    """Create a desktop icon to launch the rename-replays script."""
    if platform.system() != "Windows":
        logger.error("This subcommand is only currently available on Windows.")
    else:
        batch = (
            Path(os.environ["USERPROFILE"]) / "Desktop" / "Rename Stormgate Replays.bat"
        )
        with batch.open("wt") as f:
            print(
                """
@echo off
python -m shroudstone rename-replays
echo Press any key to close this window.
pause >nul
""",
                file=f,
            )
        logger.info(f"Batch file created at {batch} - should be visible on your desktop :)")


@app.command()
def rename_replays(
    replay_dir: Annotated[
        Optional[Path],
        typer.Option(file_okay=False, dir_okay=True, exists=True, readable=True),
    ] = None,
    my_player_id: Optional[str] = None,
    backup: bool = True,
    dry_run: bool = False,
    reprocess: Annotated[
        bool, typer.Option(help="Reprocess old replays that have already been renamed")
    ] = False,
    check_nicknames: Annotated[
        bool,
        typer.Option(
            help="Check that nicknames in the replay match those from stormgateworld"
        ),
    ] = True,
):
    """Automatically rename your replay files (including player nicknames, races, map name, duration, result)"""
    config = Config.load()
    if replay_dir is None:
        replay_dir = get_replay_dir(config)
    if my_player_id is None:
        my_player_id = get_player_id(config)
    renamer.rename_replays(
        replay_dir=replay_dir,
        my_player_id=my_player_id,
        dry_run=dry_run,
        backup=backup,
        reprocess=reprocess,
        check_nicknames=check_nicknames,
    )


def get_player_id(config: Config) -> str:
    if config.my_player_id is None:
        typer.echo(
            "You have not yet configured your player ID. To find it:\n"
            "1. visit https://stormgateworld.com/leaderboards/ranked_1v1 and search for your in-game nickname.\n"
            "2. find your account in the results and click on it.\n"
            "3. click the characters next to the '#' icon to copy your player ID.\n"
            "4. paste it below and press enter :)"
        )
        config.my_player_id = typer.prompt("Player ID")
    config.save()
    return config.my_player_id


def guess_replay_dir() -> Optional[Path]:
    if platform.system() == "Windows":
        # Should be easy, just look in the current user's local app data
        appdata = os.environ["LOCALAPPDATA"]
        paths = [Path(appdata) / "Stormgate" / "Saved" / "Replays"]
    else:
        # If this script is running on Linux and Stormgate is installed using
        # Steam+Proton, look in the steam compatdata:
        steammnt = (
            Path("~").expanduser()
            / ".steam/root/steamapps/compatdata/2012510/pfx/dosdevices"
        )

        # If this script is running on the Windows Subsystem for Linux, we can
        # find the Windows drives mounted in /mnt:
        wslmnt = Path("/mnt")

        tail = "AppData/Local/Stormgate/Saved/Replays"
        paths = [
            *steammnt.glob("*/users/steamuser/{tail}"),
            *wslmnt.glob(f"*/Users/*/{tail}"),
            *wslmnt.glob(f"*/Documents and Settings/*/{tail}"),
        ]

    for path in paths:
        if path.is_dir():
            return path


def get_replay_dir(config: Config) -> Path:
    if config.replay_dir is None:
        config.replay_dir = guess_replay_dir()
        if config.replay_dir is None:
            typer.echo("Couldn't automatically determine your replay directory!")
            typer.echo(
                "It normally lives in your user directory under "
                r"AppData\Local\Stormgate\Saved\Replays."
            )
            typer.echo("If you know the correct path, please enter it now.")
            config.replay_dir = Path(typer.prompt("Path to replay directory"))
        config.save()
    return config.replay_dir



"""Persistent configuration stored in standard user data directories"""
import os
import platform
from pathlib import Path
from typing import Optional

from pydantic import BaseModel


def _platform_data_dir() -> Path:
    if platform.system() == "Windows":
        return Path(os.environ["LOCALAPPDATA"])
    else:
        xdg_data_home = os.environ.get("XDG_DATA_HOME")
        if xdg_data_home:
            return Path(xdg_data_home)
        return Path("~").expanduser() / ".local" / "share"


data_dir = _platform_data_dir() / "shroudstone"
data_dir.mkdir(parents=True, exist_ok=True)
config_file = data_dir / "config.json"


class Config(BaseModel):
    my_player_id: Optional[str] = None
    replay_dir: Optional[Path] = None

    @staticmethod
    def load():
        if config_file.exists():
            return Config.model_validate_json(config_file.read_text())
        else:
            return Config()

    def save(self):
        config_file.write_text(self.model_dump_json(indent=2))
